from celery import Celery, signals

from cadasta.workertoolbox.conf import Config
from cadasta.workertoolbox.tests import build_functional_tests


app = Celery()
conf = Config(imports=tuple())
app.config_from_object(conf)


class FunctionalTests(build_functional_tests(app)):

    def test_set_chord_unlock_retry_limit(self):
        """
        Ensure that CHORD_UNLOCK_MAX_RETRIES are applied when worker
        signal is sent.
        """
        my_app = Celery()
        my_app.config_from_object(Config(CHORD_UNLOCK_MAX_RETRIES=1234))

        self.assertEqual(my_app.tasks['celery.chord_unlock'].max_retries, None)

        class sender():
            app = my_app

        signals.worker_init.send(sender=sender)
        self.assertEqual(my_app.tasks['celery.chord_unlock'].max_retries, 1234)



import unittest
from unittest.mock import MagicMock

from cadasta.workertoolbox import utils


class TestUtils(unittest.TestCase):
    def test_extract_followups(self):
        request = MagicMock(
            callbacks=['1', '2', '3'],
            errbacks=['a', 'b', 'c'],
        )
        mock_task = MagicMock(request=request)

        output = utils.extract_followups(mock_task)

        # Assert links and link_errors were extracted
        self.assertEqual(
            output,
            {
                'link': ['1', '2', '3'],
                'link_error': ['a', 'b', 'c'],
            }
        )
        # Assert links were removed from request
        self.assertEqual(mock_task.request.callbacks, None)



def extract_followups(task):
    """
    Retrieve callbacks and errbacks from provided task instance, disables
    tasks callbacks.
    """
    callbacks = task.request.callbacks
    errbacks = task.request.errbacks
    task.request.callbacks = None
    return {'link': callbacks, 'link_error': errbacks}



import unittest
from mock import patch, MagicMock


def build_functional_tests(app, is_worker=True):
    """
    Helper to produce sanity tests for provided configuration.

    app: A configured Celery app instance
    is_worker: If this codebase is to be run as a Celery worker (as
        opposed to a codebase that is only a task producer).
    """
    class TestConfigFunctional(unittest.TestCase):
        """ Ensure that our configuration provides expected functionality """

        @classmethod
        @patch('kombu.transport.SQS.Channel.sqs', MagicMock())
        def setUpClass(cls):
            cls.app = app
            if is_worker:
                app.Worker()  # Init worker (sends signal)
            cls.channel = cls.app.connection().channel()

        def test_default_queue_name(self):
            """ Ensure default queue is correctly named """
            self.assertEqual(
                self.app.conf.task_default_queue, 'celery')

        def test_platform_queue_name(self):
            """ Ensure platform queue is correctly named """
            self.assertEqual(
                self.app.conf.PLATFORM_QUEUE_NAME, 'platform.fifo')

        def test_default_exchange_type(self):
            """ Ensure default exchange is topic exchange """
            def_exch = self.app.conf.task_default_exchange
            exch_type = self.channel.typeof(def_exch).type
            self.assertEqual(exch_type, 'topic')

        def test_default_exchange_routing(self):
            """ Ensure default exchange routes tasks to multiple queues """
            exchange = self.app.conf.task_default_exchange
            for q in self.app.conf.QUEUES:
                queues = self.channel.typeof(exchange).lookup(
                    table=self.channel.get_table(exchange),
                    exchange=exchange, routing_key=q,
                    default=self.app.conf.task_default_queue)
                self.assertEqual(len(queues), 2)
                self.assertTrue(q in queues)
                self.assertTrue(self.app.conf.PLATFORM_QUEUE_NAME in queues)

        def test_celery_exchange_routing(self):
            """
            Ensure celery queue and platform queue are registered with default
            exchange
            """
            exchange = self.app.conf.task_default_exchange
            queues = self.channel.typeof(exchange).lookup(
                table=self.channel.get_table(exchange),
                exchange=exchange, routing_key='celery',
                default=self.app.conf.task_default_queue)

            self.assertEqual(len(queues), 2)
            self.assertTrue('celery' in queues)
            self.assertTrue(self.app.conf.PLATFORM_QUEUE_NAME in queues)

        def test_celery_task_routing(self):
            """
            Ensure celery tasks route to celery queue and platform queue
            """
            options = self.app.amqp.router.route({}, 'celery.chord_unlock')
            self.assertNotIn('queue', options)
            self.assertIn('exchange', options)
            self.assertIn('routing_key', options)
            exchange = options['exchange'].name
            routing_key = options['routing_key']

            queues = self.channel.typeof(exchange).lookup(
                table=self.channel.get_table(exchange),
                exchange=exchange, routing_key=routing_key,
                default=self.app.conf.task_default_queue)
            self.assertEqual(len(queues), 2)
            self.assertTrue('celery' in queues)
            self.assertTrue(self.app.conf.PLATFORM_QUEUE_NAME in queues)

        def test_max_retries(self):
            """ Ensure that, by default, max_retries is set to an int """
            self.assertEqual(
                type(self.app.tasks['celery.chord_unlock'].max_retries),
                int)

    return TestConfigFunctional



from celery.signals import worker_init


@worker_init.connect
def setup_exchanges(sender, **kwargs):
    """ Setup result exchange to route all tasks to platform queue """
    app = sender.app
    with app.producer_or_acquire() as P:
        # Ensure all queues are noticed and configured with their
        # appropriate exchange.
        for q in app.amqp.queues.values():
            P.maybe_declare(q)


@worker_init.connect
def limit_chord_unlock_tasks(sender, **kwargs):
    """
    Set max_retries for chord.unlock tasks to avoid infinitely looping
    tasks. (see celery/celery#1700 or celery/celery#2725)
    """
    task = sender.app.tasks['celery.chord_unlock']
    if task.max_retries is None:
        retries = getattr(sender.app.conf, 'CHORD_UNLOCK_MAX_RETRIES', None)
        task.max_retries = retries



import codecs
import os
import re
import sys
from setuptools import setup, find_packages, Command


###
NAME = 'cadasta-workertoolbox'
META_PATH = ["cadasta", "workertoolbox", "__init__.py"]
PACKAGES = find_packages()
CLASSIFIERS = [
    "Development Status :: 3 - Alpha",
]
REQUIREMENTS = [
    'boto3>=1.4.4,<=1.5',
    'celery>=4.1.0,<=4.2',
    'kombu>=4.1.0,<=4.2',
    'psycopg2>=2.7.1',
    'SQLAlchemy>=1.1.11,<=1.2',
    'pycurl>=7.43.0,<=7.44',
    'mock>=2.0.0',
]
###


def read(parts):
    """
    Build an absolute path from parts array and and return the contents
    of the resulting file.  Assume UTF-8 encoding.
    """
    cur_dir = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(cur_dir, *parts), "rb", "utf-8") as f:
        return f.read()


META_FILE = read(META_PATH)


def find_meta(meta):
    """
    Extract __*meta*__ from META_FILE.
    """
    meta_match = re.search(
        r"^__{meta}__ = ['\"]([^'\"]*)['\"]".format(meta=meta),
        META_FILE, re.M
    )
    if meta_match:
        return meta_match.group(1)
    raise RuntimeError("Unable to find __{meta}__ string.".format(meta=meta))


def ensure_clean_git(operation='operation'):
    """ Verify that git has no uncommitted changes """
    if os.system('git diff-index --quiet HEAD --'):
        print("Unstaged or uncommitted changes detected. {} aborted.".format(
            operation.capitalize()))
        sys.exit()


class CleanCommand(Command):
    """ Custom clean command to tidy up the project root. """
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        os.system('rm -vrf ./build ./dist ./*.pyc ./*.tgz ./*.egg-info')


class PublishCommand(Command):
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        ensure_clean_git('publishing')
        if os.system("pip freeze | grep twine"):
            print("twine not installed.\nUse `pip install twine`.\nExiting.")
            sys.exit()
        os.system("python setup.py sdist")
        os.system("python setup.py bdist_wheel")
        os.system("twine upload --skip-existing dist/*")


class TagCommand(Command):
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        ensure_clean_git('tagging')
        os.system("git tag -a {0} -m 'version {0}'".format(
            find_meta('version')))
        os.system("git push --tags")


class TestCommand(Command):
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        if os.system("./runtests"):
            print("Tests failed, exiting.")
            sys.exit()


if __name__ == "__main__":
    try:
        long_description = open('README.md').read()
    except IOError:
        if 'publish' in sys.argv:
            raise
        long_description = "Failed to open README.md"
    try:
        import pypandoc
        pattern = re.compile('<.*\w*>')
        no_html_descr = re.sub(pattern, '', long_description)
        long_description = pypandoc.convert_text(
            no_html_descr, 'rst', 'markdown')
    except(IOError, ImportError):
        if 'publish' in sys.argv:
            raise

    setup(
        name=NAME,
        packages=PACKAGES,
        classifiers=CLASSIFIERS,

        version=find_meta('version'),
        description=find_meta('description'),
        author=find_meta('author'),
        author_email=find_meta('email'),
        license=find_meta('license'),
        url=find_meta('url'),

        long_description=long_description,

        install_requires=REQUIREMENTS,

        cmdclass={
            'clean': CleanCommand,
            'publish': PublishCommand,
            'tag': TagCommand,
            'test': TestCommand,
        }
    )



__version__ = '0.1.13'
__license__ = 'GNU Affero General Public License v3.0'
__description__ = 'Cadasta Worker Toolbox'
__author__ = 'Anthony Lukach'
__email__ = 'alukach@cadasta.org'
__url__ = 'https://github.com/Cadasta/cadasta-workertoolbox'


DEFAULT_QUEUES = (
    'msg',
    'export',
)



import os
import pprint

from kombu import Exchange, Queue

# Ensure signals are imported before app starts
from .signals import *  # NOQA
from . import DEFAULT_QUEUES


class Config:
    # Broker
    QUEUE_NAME_PREFIX = os.environ.get('QUEUE_PREFIX', 'dev')
    broker_transport = 'sqs'
    broker_transport_options = {
        'region': 'us-west-2',
        'queue_name_prefix': '{}-'.format(QUEUE_NAME_PREFIX)
    }

    # Results
    RESULT_DB_USER = os.environ.get('RESULT_DB_USER', 'cadasta')
    RESULT_DB_PASS = os.environ.get('RESULT_DB_PASS', 'cadasta')
    RESULT_DB_HOST = os.environ.get('RESULT_DB_HOST', 'localhost')
    RESULT_DB_NAME = os.environ.get('RESULT_DB_NAME', 'cadasta')
    RESULT_DB_PORT = os.environ.get('RESULT_DB_PORT', '5432')
    result_backend = (
        'db+postgresql://{0.RESULT_DB_USER}:{0.RESULT_DB_PASS}@'
        '{0.RESULT_DB_HOST}:{0.RESULT_DB_PORT}/{0.RESULT_DB_NAME}')
    task_track_started = True

    # Exchanges
    task_default_exchange = 'task_exchange'
    task_default_exchange_type = 'topic'

    # Queues
    PLATFORM_QUEUE_NAME = 'platform.fifo'

    # Tasks
    CHORD_UNLOCK_MAX_RETRIES = 60 * 60 * 6  # 6hrs

    def __init__(self, QUEUES=DEFAULT_QUEUES, imports=('app.tasks',), **kw):
        """
        Object to manage Celery application configuration.
        """
        self.QUEUES = QUEUES
        self.imports = imports

        for k, v in kw.items():
            setattr(self, k, v)

        try:
            self.result_backend = self.result_backend.format(self)
        except ValueError:
            raise ValueError(
                "Unable to render 'result_backend' value: %r" %
                self.result_backend)

        if not hasattr(self, 'task_queues'):
            self.task_queues = self._generate_queues(
                self.QUEUES, self._default_exchange_obj,
                self.PLATFORM_QUEUE_NAME)

        if not hasattr(self, 'task_routes'):
            self.task_routes = self._route_task

    def __repr__(self):
        attr_str = pprint.pformat(self.to_dict())
        return '{0.__class__.__name__}({1})'.format(self, attr_str)

    def to_dict(self):
        return {
            k: getattr(self, k)
            for k in dir(self)
            if (k.islower() and not k.startswith('_') and
                not callable(getattr(self, k)))
        }

    @property
    def _default_exchange_obj(self):
        return Exchange(
            self.task_default_exchange,
            self.task_default_exchange_type)

    @staticmethod
    def _generate_queues(queues, exchange, platform_queue):
        """ Queues known by this worker """
        return set([
            Queue('celery', exchange, routing_key='celery'),
            Queue(platform_queue, exchange, routing_key='#'),
        ] + [
            Queue(q_name, exchange, routing_key=q_name)
            for q_name in queues
        ])

    def _route_task(self, name, args, kwargs, options, task=None, **kw):
        return {
            'routing_key': name.split('.')[0],
            'exchange': self._default_exchange_obj
        }



import unittest

from cadasta.workertoolbox.conf import Config


class TestConfigClass(unittest.TestCase):

    def test_override(self):
        """ Ensure default params can be overridden """
        conf = Config(broker_transport='foo')
        self.assertEqual(conf.broker_transport, 'foo')

    def test_repr(self):
        conf = Config(a=1, b=2, z=3)
        self.assertTrue(repr(conf).startswith("Config({'a': 1,\n 'b': 2,\n"))
        self.assertTrue(repr(conf).endswith(" 'z': 3})"))

    def test_queues(self):
        self.assertEqual(
            Config(QUEUES=['foo']).QUEUES,
            ['foo'])
        self.assertEqual(
            Config(['bar']).QUEUES,
            ['bar'])

    def test_backend_default(self):
        self.assertEqual(
            Config(QUEUES=[]).result_backend,
            'db+postgresql://cadasta:cadasta@localhost:5432/cadasta')

    def test_backend_custom(self):
        self.assertEqual(
            Config(result_backend=':memory:').result_backend,
            ':memory:')

    def test_backend_custom_rendered(self):
        self.assertEqual(
            Config(
                result_backend='{0.foo}:{0.BAR}',
                foo='abc', BAR='def'
            ).result_backend,
            'abc:def'
        )

    def test_backend_custom_error(self):
        with self.assertRaises(ValueError) as context:
            Config(
                result_backend='{0.foo}:{}',
                foo='abc'
            )
        self.assertEqual(
            getattr(context.exception, 'message', context.exception.args[0]),
            "Unable to render 'result_backend' value: '{0.foo}:{}'"
        )

    def test_default_chord_unlock_max_retries(self):
        conf = Config()
        self.assertTrue(isinstance(conf.CHORD_UNLOCK_MAX_RETRIES, int))
